/* -*-c-*-
 *
 * $Id: f-prime.c,v 1.3 2003/05/15 23:25:59 mdw Exp $
 *
 * Prime fields with Montgomery arithmetic
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: f-prime.c,v $
 * Revision 1.3  2003/05/15 23:25:59  mdw
 * Make elliptic curve stuff build.
 *
 * Revision 1.2  2002/01/13 13:48:44  mdw
 * Further progress.
 *
 * Revision 1.1  2001/04/29 18:12:33  mdw
 * Prototype version.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/sub.h>

#include "field.h"
#include "mpmont.h"

/*----- Data structures ---------------------------------------------------*/

typedef struct fctx {
  field f;
  mpmont mm;
} fctx;

/*----- Main code ---------------------------------------------------------*/

/* --- Field operations --- */

static void fdestroy(field *ff)
{
  fctx *f = (fctx *)ff;
  mpmont_destroy(&f->mm);
  DESTROY(f);
}

static mp *fin(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  return (mpmont_mul(&f->mm, d, x, f->mm.r2));
}

static mp *fout(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  return (mpmont_reduce(&f->mm, d, x));
}

static mp *fneg(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  return (mp_sub(d, f->mm.m, x));
}

static mp *fadd(field *ff, mp *d, mp *x, mp *y)
{
  return (mp_add(d, x, y));
}

static mp *fsub(field *ff, mp *d, mp *x, mp *y)
{
  return (mp_sub(d, x, y));
}

static mp *fmul(field *ff, mp *d, mp *x, mp *y)
{
  fctx *f = (fctx *)ff;
  return (mpmont_mul(&f->mm, d, x, y));
}

static mp *fsqr(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  d = mp_sqr(d, x);
  return (mpmont_reduce(&f->mm, d, d));
}

static mp *finv(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  d = mpmont_reduce(&f->mm, d, x);
  mp_gcd(0, 0, &d, f->mm.m, d);
  return (mpmont_mul(&f->mm, d, d, f->mm.r2));
}

static mp *freduce(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  mp_div(0, &d, x, f->mm.m);
  return (d);
}

static mp *fdbl(field *ff, mp *d, mp *x)
{
/*   fctx *f = (fctx *)ff; */
  return (mp_lsl(d, x, 1));
}

static mp *ftpl(field *ff, mp *d, mp *x)
{
/*   fctx *f = (fctx *)ff; */
  MP_DEST(d, MP_LEN(x) + 1, x->f);
  MPX_UMULN(d->v, d->vl, x->v, x->vl, 3);
  return (d);
}

static mp *fsqrt(field *ff, mp *d, mp *x)
{
  fctx *f = (fctx *)ff;
  d = mpmont_reduce(&f->mm, d, x);
  d = mp_modsqrt(d, d, f->mm.m);
  return (mpmont_mul(&f->mm, d, d, f->mm.r2));
}

/* --- Field operations table --- */

static field_ops fops = {
  fdestroy,
  fin, fout,
  fneg, fadd, fsub, fmul, fsqr, finv, freduce,
  fdbl, ftpl, fsqrt
};

/* --- @field_prime@ --- *
 *
 * Arguments:	@mp *p@ = the characteristic of the field
 *
 * Returns:	A pointer to the field.
 *
 * Use:		Creates a field structure for a prime field of size %$p$%,
 *		using Montgomery reduction for arithmetic.
 */

field *field_prime(mp *p)
{
  fctx *f = CREATE(fctx);
  f->f.ops = &fops;
  mpmont_create(&f->mm, p);
  f->f.zero = MP_ZERO;
  f->f.one = f->mm.r;
  return (&f->f);
}

/*----- That's all, folks -------------------------------------------------*/
